﻿namespace Serial_Port
{
    using System;
    using System.IO;
    using System.IO.Ports;
    using System.Runtime.InteropServices;
    using Microsoft.Win32.SafeHandles;
    using System.Threading;
    using System.Windows.Forms;
    using System.Drawing;

    public class PortAccess
    {
        public static SerialPort Boca_Serial;
        public static System.IO.Ports.Parity parity = System.IO.Ports.Parity.None;
        public static System.IO.Ports.StopBits stopbits = System.IO.Ports.StopBits.One;
        public static int databits = 8;
        public static int baud = 9600;
        
        //thread for read operations
        //protected Thread dataReadingThread;
        public static PortAccess boca_com;
        private boca_sample.FrmBoca m_parent;

        public PortAccess(boca_sample.FrmBoca frm1)
        {

            //create Read Thread
            //dataReadingThread = new Thread(new ThreadStart(readDataThread));
            m_parent = (boca_sample.FrmBoca)frm1;
            boca_com = this;
        }

        //7.0.0.0 added this
        public static bool Open_Serial(string portname, string b, string p, string db)
        {
            bool status = false;        //init to not found

            //if the port has not been opened before then Boca_Serial needs to be created
            if (Boca_Serial == null)
                Boca_Serial = new SerialPort();

            //6.4.0.0
            try
            {
                //if the port is already open, close it so it can be open with the current settings
                if (Boca_Serial.IsOpen)
                    Boca_Serial.Close();
            }
            catch (Exception e)                                         //6.4.0.0
            {
                MessageBox.Show("Open_Serial Close port handle invalid while opening Serial Port " + portname);
                MessageBox.Show("Exception Message = " + e.Message);    //6.4.0.0
            }
            //6.4.0.0

            stopbits = StopBits.One;
            baud = Convert.ToInt32(b);
            databits = Convert.ToInt32(db);
            if (p == "N")
                parity = Parity.None;
            else if (p == "E")
                parity = Parity.Even;
            else
                parity = Parity.Odd;

            //attempt to open
            try                                                         //6.2.1.0
            {
                Boca_Serial.BaudRate = baud;
                Boca_Serial.Parity = parity;
                Boca_Serial.StopBits = stopbits;
                Boca_Serial.DataBits = databits;
                Boca_Serial.PortName = portname;
                Boca_Serial.Open();
            }
            catch (Exception e)                                         //6.2.1.0
            {
                MessageBox.Show("Unable to open Serial Port " + portname + ". Exception Message = " + e.Message);
                //7.1.0.0 MessageBox.Show("Exception Message = " + e.Message);    //6.2.1.0
                status = false;                                         //reset to failure
            }

            if (Boca_Serial.IsOpen)
            {
                status = true;
                boca_com.m_parent.display_baud_rate();
                boca_sample.FrmBoca.checking_for_data = true;           //6.2.1.0
                //if auto negotiate is no
                if (boca_com.m_parent.AutoNo.Checked)           //1.9.0.0
                {
                    //manually set so set up event handlers
                    Boca_Serial.DataReceived += new SerialDataReceivedEventHandler(DataReceivedHandler);
                    Boca_Serial.ErrorReceived += new SerialErrorReceivedEventHandler(ErrorReceivedHandler);
                    Boca_Serial.PinChanged += new SerialPinChangedEventHandler(PinChangedHandler);
                    Boca_Serial.DtrEnable = true;           //2.7.1.0
                    Boca_Serial.RtsEnable = true;           //2.7.1.0
                    light_init();                           //2.7.1.0
                }
            }
            else
            {
                boca_com.m_parent.printer_port_closed();
                boca_com.m_parent.Display_Text("Failed to open serial port " + boca_com.m_parent.Current_Port_Name);
            }

            return (status);
        }
        //This routine will attempt to determine the communication rates of the printer and set the PC COM port to match them.
        public static bool Find_Baud(string portname)
        {

            Boca_Serial = new SerialPort();
           //System.IO.Ports.SerialPort Boca_Serial = new System.IO.Ports.SerialPort();
            int i, Max_Choices = 27;
            char ichar;                 //incoming character
            bool status = false;        //init to not found
            string incoming = "";       //incoming string of data

            if (Boca_Serial.IsOpen)
                Boca_Serial.Close();

            stopbits = StopBits.One;
            for (i = 1; i <= Max_Choices; i++)
            {

                switch (i)
                {
                    case 1:
                        baud = 9600;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 2:
                        baud = 9600;
                        parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 3:
                        baud = 9600;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 4:
                        baud = 115200;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 5:
                        baud = 115200;
                        parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 6:
                        baud = 115200;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 7:
                        baud = 1200;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 8:
                        baud = 1200;
                        parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 9:
                        baud = 1200;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 10:
                        baud = 2400;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 11:
                        baud = 2400;
                        parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 12:
                        baud = 2400;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 13:
                        baud = 4800;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 14:
                        baud = 4800;
                       parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 15:
                        baud = 4800;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 16:
                        baud = 19200;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 17:
                        baud = 19200;
                       parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 18:
                        baud = 19200;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 19:
                        baud = 28800;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 20:
                        baud = 28800;
                        parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 21:
                        baud = 28800;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 22:
                        baud = 38400;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 23:
                        baud = 38400;
                        parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 24:
                        baud = 38400;
                        parity = Parity.None;
                        databits = 8;
                        break;
                    case 25:
                        baud = 57600;
                        parity = Parity.Even;
                        databits = 7;
                        break;
                    case 26:
                        baud = 57600;
                        parity = Parity.Odd;
                        databits = 7;
                        break;
                    case 27:
                        baud = 57600;
                        parity = Parity.None;
                        databits = 8;
                        break;

                    //this is the default factory setting
                    default:
                        baud = 9600;
                        parity = Parity.None;
                        databits = 8;
                        break;
                }   //switch
                try
                {
                    Boca_Serial.BaudRate = baud;
                    Boca_Serial.Parity = parity;
                    Boca_Serial.StopBits = stopbits;
                    Boca_Serial.DataBits = databits;
                    Boca_Serial.PortName = portname;
                    Boca_Serial.Open();
                    Boca_Serial.Write("<S1>");
                    Thread.Sleep(10);                                       //10 milliseconds
                    incoming = Boca_Serial.ReadExisting();
                }
                catch
                {
                    MessageBox.Show("Unable to Open Serial Port " + portname);
                    status = false;                                         //reset to failure
                    incoming = "";
                    i = Max_Choices + 1;
                }

                if (incoming.Length > 0)
                {
                    ichar = incoming[0];
                    if (ichar.Equals((char)boca_sample.FrmBoca.XON))
                    {
                        status = true;                                     //found
                        Boca_Serial.DataReceived += new SerialDataReceivedEventHandler(DataReceivedHandler);
                        Boca_Serial.ErrorReceived += new SerialErrorReceivedEventHandler(ErrorReceivedHandler);
                        Boca_Serial.PinChanged += new SerialPinChangedEventHandler(PinChangedHandler);
                        i = Max_Choices + 1;
                    }
                    else
                        Boca_Serial.Close();
                    incoming = "";
                }
                else
                    Boca_Serial.Close();

            }       //next i
            return (status);
        }           //end find_baud

        public static void light_init()
        {
            //in a Boca Systems printer the serial port pin CD is tied high to a 5 volt line perm.
            //also depending on which interface (9 pin, 25 pin, rj11/12 jack) the CD line might not be connected.
            //given that the CD light indicator will simply copy the CTS line, which is used as CD in the printer.
            if (Boca_Serial != null)                    //1.7.1.0
            {
                if (Boca_Serial.IsOpen == true)
                {
                    //6.4.0.0 On standard serial port CTS is used on the interface and CD light is a mirror image of that
                    //6.4.0.0 On virtual serial port CD is used on the interface and CTS light is a mirror image of that
                    //if (Boca_Serial.CDHolding)
                    //boca_com.m_parent.set_cd(Color.LawnGreen);
                    //else
                    //boca_com.m_parent.set_cd(Color.Red);

                    if ((Boca_Serial.CtsHolding) || (Boca_Serial.CDHolding))            //6.4.0.0
                    {
                        boca_com.m_parent.set_cd(Color.LawnGreen);
                        boca_com.m_parent.set_cts(Color.LawnGreen);
                    }
                    else
                    {
                        boca_com.m_parent.set_cts(Color.Red);
                        boca_com.m_parent.set_cd(Color.Red);
                    }

                    if (Boca_Serial.DsrHolding)
                        boca_com.m_parent.set_dsr(Color.LawnGreen);
                    else
                        boca_com.m_parent.set_dsr(Color.Red);
                    if (Boca_Serial.DtrEnable)
                        boca_com.m_parent.set_dtr(Color.LawnGreen);
                    else
                        boca_com.m_parent.set_dtr(Color.Red);
                    if (Boca_Serial.RtsEnable)
                        boca_com.m_parent.set_rts(Color.LawnGreen);
                    else
                        boca_com.m_parent.set_rts(Color.Red);
                    boca_com.m_parent.set_td(Color.Red);
                    boca_com.m_parent.set_rd(Color.Red);
                }
            }

        }

        public static void lights_off()
        {
            boca_com.m_parent.set_cts(Color.Red);
            boca_com.m_parent.set_cd(Color.Red);
            boca_com.m_parent.set_dsr(Color.Red);
            boca_com.m_parent.set_dtr(Color.Red);
            boca_com.m_parent.set_rts(Color.Red);
            boca_com.m_parent.set_td(Color.Red);
            boca_com.m_parent.set_rd(Color.Red);
        }

        public static void write_serial(string command)
        {
            bool WriteComplete = false;                                                 //7.0.0.0

            if (Boca_Serial != null)                                                    //1.7.1.0
            {

                //6.4.0.0 
                try
                {
                    while (WriteComplete == false)                                      //7.0.0.0
                    {
                        if (Boca_Serial.IsOpen)
                        {
                            //8.2.0.0 Check Data Set Ready before transmission unless BT
                            if ((Boca_Serial.DsrHolding) || (boca_com.m_parent.BT_Status)) //7.0.0.0   
                            {
                                boca_com.m_parent.set_td(Color.LawnGreen);
                                boca_com.m_parent.set_rd(Color.Red);
                                boca_com.m_parent.Refresh();
                                Boca_Serial.Write(command);
                                WriteComplete = true;                                   //7.0.0.0
                                boca_com.m_parent.set_td(Color.Red);
                            }
                        }
                        else
                        {
                            MessageBox.Show("Error " + Boca_Serial.PortName + " port is closed so write is incomplete.");
                            WriteComplete = true;                                           //7.0.0.0 exit while loop
                        }
                    }
                }
                catch (Exception e)                                                         //6.4.0.0
                {
                    MessageBox.Show("Write Failed, handle invalid on current Port " + Boca_Serial.PortName + ". Close and then reopen the port" + " Exception Message = " + e.Message);
                }
                //6.4.0.0
            }
        }

        public static void write_serialb(byte[] buffer)
        {
            int offset = 0;                                                             //7.0.0.0
            int count = 2048;                                                           //7.0.0.0

            if (Boca_Serial != null)
            {
                //6.4.0.0 
                try
                {
                    while (offset < buffer.Length)
                    {
                        if (Boca_Serial.IsOpen)
                        {
                            if (Boca_Serial.DsrHolding)                                     //7.0.0.0   Check Data Set Ready before next transmission
                            {
                                boca_com.m_parent.set_td(Color.LawnGreen);
                                boca_com.m_parent.set_rd(Color.Red);
                                boca_com.m_parent.Refresh();

                                //test total left for last write
                                if ((buffer.Length - offset) < count)                       //7.0.0.0
                                    count = buffer.Length - offset;                         //7.0.0.0

                                //Boca_Serial.Write(buffer, 0, buffer.Length);              //7.0.0.0 buffer, offset, count
                                Boca_Serial.Write(buffer, offset, count);                   //7.0.0.0 buffer, offset, count
                                offset += count;                                            //7.0.0.0
                                boca_com.m_parent.set_td(Color.Red);
                            }
                        }
                        else
                        {
                            MessageBox.Show("Error " + Boca_Serial.PortName + " port is closed so write is incomplete.");
                            offset = buffer.Length + 1;          // exit while loop
                        }
                    }
                }
                catch (Exception e)                                         //6.4.0.0
                {
                    MessageBox.Show("Write Failed, handle invalid on current Port " + Boca_Serial.PortName + ". Close and then reopen the port" + " Exception Message = " + e.Message);
                }
                //6.4.0.0
            }
        }

        //since read_serial and DataReceivedHandler are used in the read thread they cannot directly change the
        //serial port lights on the main form.  Instead a callback funtion DisplayLight will be used.
        //
        //  Value   Meaning
        //  =====   ==============
        //   0       CD  - Green
        //   1       CD  - Red
        //   2       CTS - Green
        //   3       CTS - Red
        //   4       DSR - Green
        //   5       DSR - Red
        //   6       DTR - Green
        //   7       DTR - Red
        //   8       RTS - Green
        //   9       RTS - Red
        //  10       TD  - Green
        //  11       TD  - Red
        //  12       RD  - Green
        //  13       RD  - Red
        //
        public static string read_serial()
        {
            boca_com.m_parent.DisplayLight(11);         //set_td(Color.Red)
            boca_com.m_parent.DisplayLight(12);         //set_rd(Color.LawnGreen)
            string incoming = Boca_Serial.ReadExisting();
            boca_com.m_parent.DisplayLight(13);         //set_rd(Color.Red)
            return (incoming);
        }

        private static void DataReceivedHandler(object sender, SerialDataReceivedEventArgs e)
        {
            boca_com.m_parent.DisplayLight(11);         //set_td(Color.Red)
            boca_com.m_parent.DisplayLight(12);         //set_rd(Color.LawnGreen)
            boca_com.m_parent.incoming = Boca_Serial.ReadExisting();
            boca_com.m_parent.Establish_Status();
            boca_com.m_parent.DisplayLight(13);         //set_rd(Color.Red)
        }

        //PinChangeHandler will use the callback function DisplayLight 
        //as a means to toggle the lights on the main form.  An integer
        //value will be passed to control the seven lights and two colors
        //
        //  Value   Meaning
        //  =====   ==============
        //   0       CD  - Green
        //   1       CD  - Red
        //   2       CTS - Green
        //   3       CTS - Red
        //   4       DSR - Green
        //   5       DSR - Red
        //   6       DTR - Green
        //   7       DTR - Red
        //   8       RTS - Green
        //   9       RTS - Red
        //  10       TD  - Green
        //  11       TD  - Red
        //  12       RD  - Green
        //  13       RD  - Red
        //
        private static void PinChangedHandler(object sender,SerialPinChangedEventArgs e)
        {
            int indicator=14;
            //in a Boca Systems printer the serial port pin CD is tied high to a 5 volt line perm.
            //also depending on which interface (9 pin, 25 pin, rj11/12 jack) the CD line might not be connected.
            //given that the CDlight indicator will simply copy the CTS line, which is used as CD
            switch (e.EventType)
            {
                case (SerialPinChange.CDChanged):
                    if (Boca_Serial.CDHolding)
                        indicator = 0;
                    else
                        indicator = 1;
                    break;
                case (SerialPinChange.CtsChanged):
                    if (Boca_Serial.CtsHolding)
                        indicator = 2;
                    else
                        indicator = 3;
                    break;
                case (SerialPinChange.DsrChanged):
                    if (Boca_Serial.DsrHolding)
                        indicator = 4;
                    else
                        indicator = 5;
                    break;
                    //test for break
                case (SerialPinChange.Break):
                    MessageBox.Show("Break State = " + Boca_Serial.BreakState.ToString());
                    break;
                    //test for ring indicator
                case (SerialPinChange.Ring):
                    MessageBox.Show("Ring Indicator on port = " + Boca_Serial.PortName);
                    break;
                default:
                    if (Boca_Serial.DtrEnable)
                        indicator = 6;
                    else
                        indicator = 7;

                    if (Boca_Serial.RtsEnable)
                        indicator = 8;
                    else
                        indicator = 9;
                    break;

            }
            boca_com.m_parent.DisplayLight(indicator);
            //MessageBox.Show("Serial port pin change event " + e.EventType.ToString());
        }

        private static void ErrorReceivedHandler(object sender, SerialErrorReceivedEventArgs e)
        {
            MessageBox.Show("Serial port error " + e.EventType.ToString());
        }



    }
}
