﻿namespace boca_sample
{
    using System;
    using System.IO;
    using System.IO.Ports;
    using System.Threading;
    using System.Net.Sockets;
    using System.Drawing;
    using System.Drawing.Printing;
    using System.Windows.Forms;
    using System.Runtime.InteropServices;
    using System.Diagnostics;               //For PDF

    public class RawPrinterHelper
    {
        // Structure and API declarions:
        public static IntPtr Write_Printer_Handle = new IntPtr(0);
        public static IntPtr Read_Printer_Handle = new IntPtr(0);
        public static string port_name = "";
        public static int port_access = 0;
 
        //Constants
        public const int PRINTER_ACCESS_ADMINISTER = 0x00000004;
        public const int PRINTER_ALL_ACCESS = 0xf000c;
        public const int PRINTER_USE_ACCESS = 0x8;
        public const int PRINTER_ACCESS_MANAGE_LIMITED = 0x40;

        public const int PRINTER_STATUS_READY = 0x0;
        public const int PRINTER_STATUS_PAUSED = 0x1;
        public const int PRINTER_STATUS_ERROR = 0x2;
        public const int PRINTER_STATUS_JAM = 0x8;
        public const int PRINTER_STATUS_OUT = 0x10;
        public const int PRINTER_STATUS_OFFLINE = 0x80;
        public const int PRINTER_STATUS_PRINTING = 0x400;
        public const int PRINTER_ATTRIBUTE_QUEUED = 0x1;
        public const int PRINTER_ATTRIBUTE_DIRECT = 0x2;
        public const int PRINTER_ATTRIBUTE_DEFAULT = 0x4;
        public const int PRINTER_ATTRIBUTE_SHARED = 0x8;
        public const int PRINTER_ATTRIBUTE_NETWORK = 0x10;
        public const int PRINTER_ATTRIBUTE_HIDDEN = 0x20;
        public const int PRINTER_ATTRIBUTE_LOCAL = 0x40;
        public const int PRINTER_ATTRIBUTE_ENABLEDEVQ = 0x80;
        public const int PRINTER_ATTRIBUTE_KEEPPRINTEDJOBS = 0x100;
        public const int PRINTER_ATTRIBUTE_DO_COMPLETE_FIRST = 0x200;
        public const int PRINTER_ATTRIBUTE_WORK_OFFLINE = 0x400;
        public const int PRINTER_ATTRIBUTE_ENABLE_BIDI = 0x800;
        public const int PRINTER_ATTRIBUTE_RAW_ONLY = 0x1000;
        public const int PRINTER_ATTRIBUTE_PUBLISHED = 0x2000;
        public const int PRINTER_CONTROL_PAUSE = 1;
        public const int PRINTER_CONTROL_RESUME = 2;
        public const int PRINTER_CONTROL_PURGE = 3;
        public const int PRINTER_CONTROL_SET_STATUS = 4;

        //@mdh Set Printer Driver setting 7.0.0.0
        public const int DM_ORIENTATION = 0x00000001;
        public const int DM_PAPERSIZE = 0x00000002;
        public const int DM_PAPERLENGTH = 0x00000004;
        public const int DM_PAPERWIDTH = 0x00000008;
        public const int DM_SCALE = 0x00000010;
        public const int DM_POSITION = 0x00000020;
        public const int DM_NUP = 0x00000040;
        public const int DM_DISPLAYORIENTATION = 0x00000080;
        public const int DM_COPIES = 0x00000100;
        public const int DM_DEFAULTSOURCE = 0x00000200;
        public const int DM_PRINTQUALITY = 0x00000400;
        public const int DM_COLOR = 0x00000800;
        public const int DM_DUPLEX = 0x00001000;
        public const int DM_YRESOLUTION = 0x00002000;
        public const int DM_TTOPTION = 0x00004000;
        public const int DM_COLLATE = 0x00008000;
        public const int DM_FORMNAME = 0x00010000;
        public const int DM_LOGPIXELS = 0x00020000;
        public const int DM_BITSPERPEL = 0x00040000;
        public const int DM_PELSWIDTH = 0x00080000;
        public const int DM_PELSHEIGHT = 0x00100000;
        public const int DM_DISPLAYFLAGS = 0x00200000;
        public const int DM_DISPLAYFREQUENCY = 0x00400000;
        public const int DM_ICMMETHOD = 0x00800000;
        public const int DM_ICMINTENT = 0x01000000;
        public const int DM_MEDIATYPE = 0x02000000;
        public const int DM_DITHERTYPE = 0x04000000;
        public const int DM_PANNINGWIDTH = 0x08000000;
        public const int DM_PANNINGHEIGHT = 0x10000000;
        public const int DM_DISPLAYFIXEDOUTPUT = 0x20000000;

        public const int DM_OUT_BUFFER = 2;
        public const int DM_IN_BUFFER = 8;
        public const short LETTER = 0x0001;
        public const short CONCERT = 0x0077;
        public const short RECEIPT = 0x007B;
        public const short CINEMA = 0x0078;
        public const short CREDIT = 0x0079;
        public const short SKI = 0x007A;

        public const short DMORIENT_PORTRAIT = 1;
        public const short DMORIENT_LANDSCAPE = 2;
        //@mdh Set Printer Driver setting 7.0.0.0



        //thread for read operations
        protected Thread dataReadingThread;
        public static RawPrinterHelper boca_raw;
        private static boca_sample.FrmBoca m_parent;
        public static DOCINFOA di = new DOCINFOA();

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public class PRINTERDEFAULTS
        {
            public string pDataType;
            public int pDevMode;
            public int DesiredAccess;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public class DOCINFOA
        {
            [MarshalAs(UnmanagedType.LPStr)]
            public string pDocName;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pOutputFile;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pDataType;
        }

        [StructLayout(LayoutKind.Sequential)]       //, CharSet = CharSet.Ansi
        internal struct structDevMode
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
            public String dmDeviceName;
            [MarshalAs(UnmanagedType.U2)]
            public short dmSpecVersion;
            [MarshalAs(UnmanagedType.U2)]
            public short dmDriverVersion;
            [MarshalAs(UnmanagedType.U2)]
            public short dmSize;
            [MarshalAs(UnmanagedType.U2)]
            public short dmDriverExtra;
            [MarshalAs(UnmanagedType.U4)]
            public int dmFields;
            [MarshalAs(UnmanagedType.I2)]
            public short dmOrientation;
            [MarshalAs(UnmanagedType.I2)]
            public short dmPaperSize;
            [MarshalAs(UnmanagedType.I2)]
            public short dmPaperLength;
            [MarshalAs(UnmanagedType.I2)]
            public short dmPaperWidth;
            [MarshalAs(UnmanagedType.I2)]
            public short dmScale;
            [MarshalAs(UnmanagedType.I2)]
            public short dmCopies;
            [MarshalAs(UnmanagedType.I2)]
            public short dmDefaultSource;
            [MarshalAs(UnmanagedType.I2)]
            public short dmPrintQuality;
            [MarshalAs(UnmanagedType.I2)]
            public short dmColor;
            [MarshalAs(UnmanagedType.I2)]
            public short dmDuplex;
            [MarshalAs(UnmanagedType.I2)]
            public short dmYResolution;
            [MarshalAs(UnmanagedType.I2)]
            public short dmTTOption;
            [MarshalAs(UnmanagedType.I2)]
            public short dmCollate;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
            public String dmFormName;
            [MarshalAs(UnmanagedType.U2)]
            public short dmLogPixels;
            [MarshalAs(UnmanagedType.U4)]
            public int dmBitsPerPel;
            [MarshalAs(UnmanagedType.U4)]
            public int dmPelsWidth;
            [MarshalAs(UnmanagedType.U4)]
            public int dmPelsHeight;
            [MarshalAs(UnmanagedType.U4)]
            public int dmNup;
            [MarshalAs(UnmanagedType.U4)]
            public int dmDisplayFrequency;
            [MarshalAs(UnmanagedType.U4)]
            public int dmICMMethod;
            [MarshalAs(UnmanagedType.U4)]
            public int dmICMIntent;
            [MarshalAs(UnmanagedType.U4)]
            public int dmMediaType;
            [MarshalAs(UnmanagedType.U4)]
            public int dmDitherType;
            [MarshalAs(UnmanagedType.U4)]
            public int dmReserved1;
            [MarshalAs(UnmanagedType.U4)]
            public int dmReserved2;
        }

        [DllImport("winspool.Drv", EntryPoint = "OpenPrinterA", SetLastError = true, CharSet = CharSet.Ansi, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool OpenPrinter([MarshalAs(UnmanagedType.LPStr)] string szPrinter, out IntPtr hPrinter, IntPtr pd);

        [DllImport("winspool.Drv", EntryPoint = "ClosePrinter", SetLastError = true, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool ClosePrinter(IntPtr hPrinter);

        [DllImport("winspool.Drv", EntryPoint = "StartDocPrinterA", SetLastError = true, CharSet = CharSet.Ansi, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern int StartDocPrinter(IntPtr hPrinter, Int32 level, [In, MarshalAs(UnmanagedType.LPStruct)] DOCINFOA di);

        [DllImport("winspool.Drv", EntryPoint = "EndDocPrinter", SetLastError = true, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool EndDocPrinter(IntPtr hPrinter);

        [DllImport("winspool.Drv", EntryPoint = "StartPagePrinter", SetLastError = true, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool StartPagePrinter(IntPtr hPrinter);

        [DllImport("winspool.Drv", EntryPoint = "EndPagePrinter", SetLastError = true, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool EndPagePrinter(IntPtr hPrinter);

        [DllImport("winspool.Drv", EntryPoint = "WritePrinter", SetLastError = true, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool WritePrinter(IntPtr hPrinter, IntPtr pBytes, Int32 dwCount, out Int32 dwWritten);

        [DllImport("winspool.Drv", EntryPoint = "ReadPrinter", SetLastError = true, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool ReadPrinter(IntPtr hPrinter, IntPtr pBytes, Int32 dwCount, out Int32 dwRead);

        [DllImport("winspool.Drv", EntryPoint = "GetPrinterA", SetLastError = true, CharSet = CharSet.Ansi, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool GetPrinter(IntPtr hPrinter, Int32 dwLevel, IntPtr pPrinter, Int32 dwBuf, out Int32 dwNeeded);

        [DllImport("winspool.Drv", EntryPoint = "SetPrinterA", SetLastError = true, CharSet = CharSet.Ansi, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool SetPrinter(IntPtr hPrinter, Int32 dwLevel, IntPtr pPrinter, Int32 command);

        [DllImport("winspool.drv", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern bool SetDefaultPrinter(string Name);                   //7.0.0.0

        [DllImport("winspool.Drv", EntryPoint = "DocumentPropertiesA", SetLastError = true,
         ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern int DocumentProperties(
           IntPtr hwnd,
           IntPtr hPrinter,
           [MarshalAs(UnmanagedType.LPStr)] string pDeviceName /* changed from String to string */,
           IntPtr pDevModeOutput,
           IntPtr pDevModeInput,
           int fMode
           );                                                               //7.0.0.0


        [DllImport("ddkfglmon.dll", EntryPoint = "ReceiveData", SetLastError = true, CharSet = CharSet.Ansi, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern int ReceiveData(string Port, IntPtr pBytes);

        [StructLayout(LayoutKind.Sequential)]
        private class PRINTER_INFO_2
        {
            [MarshalAs(UnmanagedType.LPStr)]
            public string pServerName;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pPrinterName;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pShareName;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pPortName;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pDriverName;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pComment;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pLocation;
            public IntPtr pDevMode;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pSepFile;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pPrintProcessor;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pDatatype;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pParameters;
            public IntPtr pSecurityDescriptor;
            public Int32 Attributes;
            public Int32 Priority;
            public Int32 DefaultPriority;
            public Int32 StartTime;
            public Int32 UntilTime;
            public Int32 Status;
            public Int32 cJobs;
            public Int32 AveragePPM;
        }

        //5.0.0.0
        [StructLayout(LayoutKind.Sequential)]
        public class PRINTER_INFO_6
        {
            public Int32 dwStatus;
        }

        //establish read thread (port monitor), parent for callback and a boca_raw instance
        public RawPrinterHelper(boca_sample.FrmBoca frm1)
        {
            //create Read Thread
            dataReadingThread = new Thread(new ThreadStart(readDataThread));
            //setup parent pointer back to main app
            m_parent = (boca_sample.FrmBoca)frm1;
            //current instance
            boca_raw = this;
        }

        //The following four main routines demonstrate the four functions needed to communicate with a Boca Systems printer
        //using a Boca Systems printer driver in "RAW" mode.
        //Open_Printer  - Opens 2 channels for communication.  Establishes globals Write_Printer_Handle, Read_Printer_Handle and port_name
        //Close_Printer - Closes all channels and clears globals
        //Write_Printer - uses Write_Printer_Handle to send raw text to printer
        //Read_Printer  - uses Read_Printer_Handle to read data returned from the printer 
        //
        //Simple put there are three ways (modes) to open channels to the printer through our printer drivers and other ways without 
        //drivers at the interface level.  All these modes are demonstrated in this sample project.  
        //For now a top level overview of the printer driver modes.  Before selecting a printer driver from the drop down
        //menu, select one of the three driver modes.  The default is Mode #2 - direct bidirectional
        //
        //FGL Driver modes
        //1. Spool unidirectional – Open write channels for printing only.  Most popular mode.  Least information, but fastest.  
        //   With spooler so printer can be shared. 
        //2. Direct bidirectional - Open both write and read channel for printing and reading status in between each ticket printed.  
        //   All information, and less delay since the spooler is circumvented.  Faster than with spooler.  Without spooler printer is dedicated to one user at a time, but shared in a limited way.  Without spooler OS variables cannot be updated and monitored.
        //3. Spool bidirectional – Open both write and read channel for printing and reading status in between each ticket printed.  
        //   All information, with slight delay between tickets. With spooler so printer can be completely shared and OS variables 
        //   can be monitored as updated. 
        //
        //open_printer will attempt to open two channels for communication. Before either channel is open the port name 
        //will be retrieved from the driver.  It is used to open the read channel.  The standard write channel will be opened first.
        //If successful the write handle will be stored in the global variable Write_Printer_Handle.  An attempt to open
        //a second channel for Read purposes.  If successful the read handle will be stored in the global variable Read_Printer_Handle.
        public static int Open_Printer(string szPrinterName)
        {

            IntPtr hPrinter = new IntPtr(0);
            //DOCINFOA di = new DOCINFOA();
            PRINTERDEFAULTS df = new PRINTERDEFAULTS();
            int x, bSuccess = 0;                          // Assume failure unless you specifically succeed.
            bool online = false;                //1.1.2.1
            char backslash = (char)92;

            //Init Document info structure
            di.pDocName = "Boca Systems";
            di.pDataType = "RAW";
            di.pOutputFile = null;

            //Init Printer Default Structure
            df.pDataType = "RAW";
            df.pDevMode = 0;

            //6.3.0.0
            if ((FrmBoca.CurrentOperatingSystem == "Windows 8/10") || (FrmBoca.CurrentOperatingSystem == "Windows 10"))           //@mdh test win 10
                df.DesiredAccess = PRINTER_ACCESS_MANAGE_LIMITED;           //@mdh test win 10
            else
                df.DesiredAccess = PRINTER_ALL_ACCESS;                      //@mdh test win 10            

            //setup int pointer to printer default structure
            IntPtr def = Marshal.AllocHGlobal(Marshal.SizeOf(df));
            Marshal.StructureToPtr(df, def, false);

            //Check for Boca PCL driver installed.  If so it needs to be in Mode #1
            //7.0.0.0
            x = szPrinterName.IndexOf("PCL", 0);
            if ((x != -1) && (m_parent.DRMode == 3))    //8.0.0.0 if HP driver and trying to set spool bidi, change to spool uni
                m_parent.DRMode = 1;

            Set_Attributes(szPrinterName, m_parent.DRMode);                                 //7.0.0.0
            FrmBoca.PrinterDriverName = szPrinterName;                                      //7.0.0.0 used for setting printer prefrences
            FrmBoca.PrinterDriverDefault = def;                                             //7.0.0.0

            //the global port_name will be initialized by calling Get_Printer_Portname.  The returned
            //string will look something like "USB001, Port" or "LPT1, Port", depending on which port is used.
            port_name = RawPrinterHelper.Get_Printer_Portname(szPrinterName, def);

            //if it is a network printer then port_name should look something like \\servername\usb007, Port
            //Problem is PRINTER_ALL_ACCESS does not work for network printers so port_name returns with \\, Port
            //If backslash is detected then change Desired Access to PRINTER_USE_ACCESS and then try again.
            x = port_name.IndexOf(backslash, 0);

            //if "\" found x will equal 0 which signafies network printer.
            //if no "\" found x will equal -1 which signifies local printer.
            if (x != -1)
            {
                //must be network printer, so change access and re-establish port_name
                df.DesiredAccess = PRINTER_USE_ACCESS;
                Marshal.StructureToPtr(df, def, false);
                port_name = RawPrinterHelper.Get_Printer_Portname(szPrinterName, def);
            }

            //Open the first channel to the printer using the printer name for the intent of writing raw data
            //if (OpenPrinter(szPrinterName.Normalize(), out Write_Printer_Handle, def))
            if (OpenPrinter(szPrinterName.Normalize(), out Write_Printer_Handle, IntPtr.Zero))
            {
                //indicate successful write access
                bSuccess = 1;

                online = Is_Printer_Online();
                if (online)
                {
                    //if successful, then Start a document to get a job_id
                    FrmBoca.job_id = StartDocPrinter(Write_Printer_Handle, 1, di);

                    //unable to read status back from network printer even though below a Read_Printer_Handle 
                    //is returned when called, so for now I have included the check for port_access and don't try to open
                    //the second port.  Maybe this will change later.  Need Read Thread to see data and right now it does not.
                    //Can only open Read thread on local printer with port_access equal PRINTER_ALL_ACCESS
                    //if network printer than port_access will equal PRINTER_USE_ACCESS
                    if ((boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver1) ||
                        (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver3) ||
                        (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Standard_Driver1))   //5.2.0.0
                    {
                        //Spooler and printer monitor in play, so indicate read access also
                        bSuccess = 2;
                    }
                    else
                    {
                        //6.3.0.0 if ((FrmBoca.job_id > 0) && (port_access == PRINTER_ALL_ACCESS))
                        if ((FrmBoca.job_id > 0) && ((port_access == PRINTER_ALL_ACCESS) || (port_access == PRINTER_ACCESS_MANAGE_LIMITED)))        //6.6.0.0
                        {
                            //When direct connect open second channel to the printer using port name for the intent of reading return status from the printer
                            //A read thread will be started using the read handle and operate as a port monitor.  When data is returned from
                            //printer the callback function DisplayText() is used to display status on the main parent form.  See ReadData() Below.
                            //if (OpenPrinter(port_name.Normalize(), out Read_Printer_Handle, def))
                            if (OpenPrinter(port_name.Normalize(), out Read_Printer_Handle, IntPtr.Zero))
                            {
                                //indicate read access also
                                bSuccess = 2;
                            }
                        }
                    }
                }
                else//1.1.2.1
                {
                    //indicate no connection because printer is offline
                    bSuccess = 3;
                }
            }
            Marshal.FreeHGlobal(def);
            return bSuccess;
        }


        //1.0.5.0 used here to exit gracefully in case a printer is suddenly offline (turned off)
        public static void quick_close()
        {

            boca_sample.FrmBoca.checking_for_data = false; //turn off loop for read thread

            if (Read_Printer_Handle.ToInt64() != 0)
                ClosePrinter(Read_Printer_Handle);              //close read channel
            if (Write_Printer_Handle.ToInt64() != 0)
            {
                EndDocPrinter(Write_Printer_Handle);            //end document
                ClosePrinter(Write_Printer_Handle);             //close write channel
            }

        }


        public static void Close_Printer()
        {
            //5.1.0.0 Int32 level = 2;
            //5.1.0.0 Int32 Needed = 0;
            IntPtr hPrinter = new IntPtr(0);
            IntPtr pPrinter = new IntPtr(0);
            PRINTER_INFO_2 pi = new PRINTER_INFO_2();
            bool online = false;
            bool bSuccess = false;                                  // Assume failure unless you specifically succeed.

            boca_sample.FrmBoca.checking_for_data = false;      // maybe needed to end read thread
            //5.0.0.0
            //stop the read thread first and close port channel
            if ((boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver1) ||
                (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver3) ||
                (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Standard_Driver1))   //5.2.0.0
            {
                //start_stop_read_thread();                   //stop read thread
                if (Write_Printer_Handle.ToInt64() != 0)
                {
                    bSuccess = EndDocPrinter(Write_Printer_Handle);        //end document
                    bSuccess = ClosePrinter(Write_Printer_Handle);         //close write channel
                }
            }
            else
            {
                if (Read_Printer_Handle.ToInt64() != 0)
                {
                    start_stop_read_thread();                   //stop read thread 
                    ClosePrinter(Read_Printer_Handle);          //close read channel
                }

                //gracefully shut down the read channel
                if (Write_Printer_Handle.ToInt64() != 0)
                {
                    //If printer has already reset due to factory config, then it is impossible to 
                    //write <S1>, so skip it
                    if (!FrmBoca.reset_flag)
                    {
                        //online = Is_Printer_Online();
                        if (online)
                        {
                            //perhaps read thread is waiting at a read statement 
                            Write_Printer("<S1>");                      //send S1 to printer to generate a one byte XOn response, freeing read
                            EndDocPrinter(Write_Printer_Handle);        //end document
                        }
                    }
                    else
                        FrmBoca.reset_flag = false;                 //reset

                    /*
                    //5.1.0.0 
                    //if the original attributes were altered restore them now
                    if (original_attributes != 0)
                    {
                        Thread.Sleep(2000);             //allow any writing to the printer to complete
                        online = Is_Printer_Online();
                        if (online)
                        {
                            // Get printer level page size needed - first call
                            if (GetPrinter(Write_Printer_Handle, level, pPrinter, 0, out Needed) == false)
                            {
                                if (Needed > 0)                 //1.0.5.0
                                {

                                    // Allocate needed memory
                                    IntPtr pBytes = Marshal.AllocCoTaskMem(Needed);

                                    // Get printer level data - second call
                                    GetPrinter(Write_Printer_Handle, level, pBytes, Needed, out Needed);

                                    // Convert printer data block into class structure
                                    Marshal.PtrToStructure(pBytes, pi);

                                    pi.Attributes = original_attributes;
                                    Marshal.StructureToPtr(pi, pBytes, false);
                                    if (!SetPrinter(Write_Printer_Handle, level, pBytes, 0))
                                    {
                                        throw new ApplicationException("Cannot reset original printer attributes " + Marshal.GetLastWin32Error());
                                    }
                                    Marshal.FreeCoTaskMem(pBytes);
                                }
                            }
                        }
                        else
                        {
                            //4.2.0.0
                            MessageBox.Show("Since the printer is offline/re-initializing, reseting the advanced driver attributes from DIRECT back to SPOOL is impossible at this time.  If necessary go to the Printer Properties screen, Advanced Tab, click on SPOOL and OK.");
                        }
                    }*/
                    //5.1.0.0 
                    ClosePrinter(Write_Printer_Handle);         //close write channel
                }
            }
            return;
        }

        //Write string to printer
        public static bool Write_Printer(string command)
        {
            bool bSuccess = false;                          // Assume failure unless you specifically succeed.
            int dwCount, dwWritten;

            //set pointer to command string
            IntPtr pBytes = Marshal.StringToCoTaskMemAnsi(command);

            //establish length of string
            dwCount = command.Length;

            //Write command to printer
            if (StartPagePrinter(Write_Printer_Handle))
            {
                // Write your bytes.
                bSuccess = WritePrinter(Write_Printer_Handle, pBytes, dwCount, out dwWritten);
                EndPagePrinter(Write_Printer_Handle);
            }

            //If dealing with a network printer the job must be closed to flush it, and then reopened in case something else comes.
            //This is also true for PCs running Windows 8 & 10.                                                  //V3.3.0.0
            //6.3.0.0 if ((port_access == PRINTER_USE_ACCESS) || (FrmBoca.CurrentOperatingSystem == "Windows 7") || (FrmBoca.CurrentOperatingSystem == "Windows 8"))     //V3.7.0.0
            //6.3.0.0 {
            EndDocPrinter(Write_Printer_Handle);        //end document
            int job_id = StartDocPrinter(Write_Printer_Handle, 1, di);
            //6.3.0.0 }

            Marshal.FreeCoTaskMem(pBytes);
            return bSuccess;
        }

        //Write from a byte array to printer
        public static bool Write_Printerb_orig(byte[] command)
        {
            bool bSuccess = false;                          // Assume failure unless you specifically succeed.
            int dwCount, dwWritten;

            //set pointer to command array
            IntPtr pBytes = Marshal.UnsafeAddrOfPinnedArrayElement(command, 0);

            //establish number of elements in array
            dwCount = command.Length;

            //Write command to printer
            if (StartPagePrinter(Write_Printer_Handle))
            {
                // Write the array to the printer
                bSuccess = WritePrinter(Write_Printer_Handle, pBytes, dwCount, out dwWritten);
                EndPagePrinter(Write_Printer_Handle);
            }

            //If dealing with a network printer the job must be closed to flush it, and then reopened in case something else comes.
            //This is also true for PCs running Windows 8.                                                  //V3.3.0.0
            if ((port_access == PRINTER_USE_ACCESS) || (FrmBoca.CurrentOperatingSystem == "Windows 7") || (FrmBoca.CurrentOperatingSystem == "Windows 8"))     //V3.7.0.0
            {
                EndDocPrinter(Write_Printer_Handle);        //end document
                int job_id = StartDocPrinter(Write_Printer_Handle, 1, di);
            }

            return bSuccess;
        }

        //Write from a byte array to printer            //5.0.0.0
        public static bool Write_Printerb(byte[] command)
        {
            bool bSuccess = false;                          // Assume failure unless you specifically succeed.
            int dwCount, dwWritten, index;
            int dwMax = 100000;
            int dwTotal = command.Length;

            index = 0;
            dwWritten = 0;

            while (dwTotal > 0)
            {
                //establish number of elements in array
                if (dwTotal < dwMax)
                    dwCount = dwTotal;
                else
                    dwCount = dwMax;
                dwTotal -= dwCount;

                //set pointer to command array
                IntPtr pBytes = Marshal.UnsafeAddrOfPinnedArrayElement(command, index);


                //Write command to printer
                if (StartPagePrinter(Write_Printer_Handle))
                {
                    // Write the array to the printer
                    bSuccess = WritePrinter(Write_Printer_Handle, pBytes, dwCount, out dwWritten);
                    //bSuccess = FlushPrinter(Write_Printer_Handle, pBytes, dwCount, out dwWritten, 0);
                    //EndPagePrinter(Write_Printer_Handle);
                }

                //set pointer to command array
                index += dwWritten;

                //If dealing with a network printer the job must be closed to flush it, and then reopened in case something else comes.
                //This is also true for PCs running Windows 8.                                                  //V3.3.0.0
                //if ((port_access == PRINTER_USE_ACCESS) || (FrmBoca.CurrentOperatingSystem == "Windows 7") || (FrmBoca.CurrentOperatingSystem == "Windows 8"))     //V3.7.0.0
                //{
                EndDocPrinter(Write_Printer_Handle);        //end document
                int job_id = StartDocPrinter(Write_Printer_Handle, 1, di);
                //}
            }

            return bSuccess;
        }

        //Read data back from printer.  In order not to hang the main application while waiting on return data
        //that may or may not show up, raeding will be performed in a seperate thread.  A call back function (delegate)
        //is used to report data back to main application.
        public static string Read_Printer()
        {
            int dwRead;
            bool bSuccess = false;                          // Assume failure unless you specifically succeed.
            string temp = "";

            //Allocate some unmanaged memory 1K for incoming bytes.
            byte[] answer = new byte[1024];
            IntPtr pAnswer = Marshal.AllocCoTaskMem(1024);
            dwRead = 0;

            //Check for whichever Boca driver is in use before attempting to read status from the printer
            if ((boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver1) ||
                (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver3) ||
                (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Standard_Driver1))   //5.2.0.0
                dwRead = ReceiveData(port_name, pAnswer);                               //5.1.0.0 Read from ddkfglmon.dll routine
            else
                bSuccess = ReadPrinter(Read_Printer_Handle, pAnswer, 1024, out dwRead); //5.1.0.0 Read directly from printer

            if (dwRead > 0)                                         //5.0.0.0
            {
                Marshal.Copy(pAnswer, answer, 0, dwRead);
                temp = System.Text.Encoding.UTF8.GetString(answer, 0, dwRead);          //5.0.0.0
                //temp = System.Text.Encoding.ASCII.GetString(answer, 0, dwRead);       //5.0.0.0
            }

            Marshal.FreeCoTaskMem(pAnswer);
            return temp;

        }

        //This routine is called by Open_Printer to perform overhead needed.
        //
        //Simple put there are three ways (modes) to open channels to the printer through our printer drivers and other ways without 
        //drivers at the interface level.  All these modes are demonstrated in this sample project.  
        //For now a top level overview of the printer driver modes.  Before selecting a printer driver from the drop down
        //menu, select one of the three driver modes.  The default is Mode #2 - direct bidirectional
        //
        //FGL Driver modes
        //1. Spool unidirectional – Open write channels for printing only.  Most popular mode.  Least information, but fastest.  
        //   With spooler so printer can be shared. 
        //2. Direct bidirectional - Open both write and read channel for printing and reading status in between each ticket printed.  
        //   All information, and less delay since the spooler is circumvented.  Faster than with spooler.  Without spooler printer is dedicated to one user at a time, but shared in a limited way.  Without spooler OS variables cannot be updated and monitored.
        //3. Spool bidirectional – Open both write and read channel for printing and reading status in between each ticket printed.  
        //   All information, with slight delay between tickets. With spooler so printer can be completely shared and OS variables 
        //   can be monitored as updated. 
        //
        //When selecting a printer driver an access level is established depending on whether the printer is installed
        //locally on the PC vs. a shared network printer.  A network printer will include a server name in it port name.
        public static string Get_Printer_Portname(string szPrinterName, IntPtr def)
        {
            string pn = "";         //Port Name
            string sn = "";         //Server Name
            int stat = 0;
            int att = 0;
            int x = 0;              //5.1.0.0 
            char backslash = (char)92;
            string ReturnValue = "";
            Int32 level = 2;
            Int32 Needed = 0;
            IntPtr hPrinter = new IntPtr(0);
            IntPtr pPrinter = new IntPtr(0);
            PRINTER_INFO_2 pi = new PRINTER_INFO_2();

            // Open the printer.
            //if (OpenPrinter(szPrinterName.Normalize(), out hPrinter, def))
            if (OpenPrinter(szPrinterName.Normalize(), out hPrinter, IntPtr.Zero))
            {
                // Get printer level page size needed - first call
                if (GetPrinter(hPrinter, level, pPrinter, 0, out Needed) == false)
                {
                    // Allocate needed memory
                    IntPtr pBytes = Marshal.AllocCoTaskMem(Needed);

                    // Get printer level data - second call
                    GetPrinter(hPrinter, level, pBytes, Needed, out Needed);

                    // Convert printer data block into class structure
                    Marshal.PtrToStructure(pBytes, pi);
                    ReturnValue = pi.pDriverName; // get printer driver name
                    sn = pi.pServerName;
                    pn = pi.pPortName;
                    stat = pi.Status;
                    att = pi.Attributes;
                    
                    //PRINTER_ATTRIBUTE_DIRECT
                    //if ((att & PRINTER_ATTRIBUTE_DIRECT) == 0)
                    //    FrmBoca.Direct = false;
                    //else
                    //    FrmBoca.Direct = true;

                    //PRINTER_ATTRIBUTE_ENABLE_BIDI
                    //if ((att & PRINTER_ATTRIBUTE_ENABLE_BIDI) == 0)
                    //    FrmBoca.Bidi = false;
                    //else
                    //    FrmBoca.Bidi = true;

                    //Check driver name to see if it is one of Boca Systems drivers or not.
                    //Boca drivers have names like "Boca FGL 26/46 300" DPI or "Boca BIDI FGL 26/46 300"
                    //or "Boca PCL 26/46 300".  Search the beginning of the name for the word Boca
                    //if "Boca" string not found x will equal -1 which signifies other (non Boca Systems) driver used.
                    x = szPrinterName.IndexOf("Boca", 0);
                    //y = szPrinterName.IndexOf("Boca PCL", 0);   //distinguish Boca PCL from Boca FGL

                    //Check server name to determine if this is a network printer 
                    //Also check for a non Boca Systems driver
                    //if either is true then use a uni-directional approach (else clause below) with no status read
                    //from the printer or through the printer monitor.
                    if ((sn == null) && (x != -1))
                    {
                        //check first attribute to determine Spool or Direct
                        if ((att & PRINTER_ATTRIBUTE_DIRECT) == 0)
                        {
                            //Spool - Now check if BIDI printer monitor needed
                            if ((att & PRINTER_ATTRIBUTE_ENABLE_BIDI) == 0)
                            {
                                //No BIDI so do not use Printer Monitor
                                //This could be FGL or PCL printer. Write with Spooler but no BIDI read
                                //MessageBox.Show("Spooler Enabled but no BIDI support needed");
                                FrmBoca.Current_Driver_Type = FrmBoca.Boca_Spooler_Driver2;
                            }
                            else
                            {
                                //BIDI support needed so use Printer Monitor to update Printer_Info_2 & 6
                                //MessageBox.Show("Spooler Enabled with BIDI support needed");
                                FrmBoca.Current_Driver_Type = FrmBoca.Boca_Spooler_Driver1;
                            }
                        }
                        else
                        {
                            //Direct - Now check if BIDI printer monitor needed
                            if ((att & PRINTER_ATTRIBUTE_ENABLE_BIDI) == 0)
                            {
                                //No BIDI so do not use Printer Monitor
                                //MessageBox.Show("Direct Connect Enabled but no BIDI support needed");
                                FrmBoca.Current_Driver_Type = FrmBoca.Boca_Standard_Driver2;
                            }
                            else
                            {
                                //BIDI support needed so use Printer Monitor to update Printer_Info_2 & 6
                                //MessageBox.Show("Direct Connect Enabled with BIDI support needed");
                                FrmBoca.Current_Driver_Type = FrmBoca.Boca_Standard_Driver1;
                            }
                        }
                        if (FrmBoca.Current_Driver_Type == FrmBoca.Boca_Spooler_Driver2)
                            port_access = PRINTER_USE_ACCESS;               //unidirectional
                        else
                        {
                            //port_access = PRINTER_ALL_ACCESS;                             //bidirectional
                            //6.3.0.0
                            if ((FrmBoca.CurrentOperatingSystem == "Windows 8/10") || (FrmBoca.CurrentOperatingSystem == "Windows 10"))           //@mdh test win 10
                                port_access = PRINTER_ACCESS_MANAGE_LIMITED;                //@mdh test win 10
                            else
                                port_access = PRINTER_ALL_ACCESS;                           //@mdh test win 10            
                        }
                    }
                    else
                    {
                        //MessageBox.Show("Network Printer or Non Boca driver.");
                        if (x == -1)                                                                //5.2.0.0 Non Boca Driver
                        {
                            FrmBoca.Current_Driver_Type = FrmBoca.Non_Boca_Driver;
                            port_access = PRINTER_USE_ACCESS;                                       //unidirectional
                        }
                        else                                                                        //5.2.0.0 Boca Network Driver
                        {
                            FrmBoca.Current_Driver_Type = FrmBoca.Boca_Spooler_Driver3;             //5.2.0.0 Monitor OS variables
                            //port_access = PRINTER_ALL_ACCESS;                             //bidirectional
                            //6.3.0.0
                            if ((FrmBoca.CurrentOperatingSystem == "Windows 8/10") || (FrmBoca.CurrentOperatingSystem == "Windows 10"))           //@mdh test win 10
                                port_access = PRINTER_ACCESS_MANAGE_LIMITED;                //@mdh test win 10
                            else
                                port_access = PRINTER_ALL_ACCESS;                           //@mdh test win 10            
                        }
                    }

                    //if in Boca_Driver with printer monitor mode the port name (pn) is set like "USB001" so just exit
                    //if not the compete the port name (pn) to include command word Port and a server name when needed
                    if ((FrmBoca.Current_Driver_Type != FrmBoca.Boca_Spooler_Driver1) &&
                        (FrmBoca.Current_Driver_Type != FrmBoca.Boca_Standard_Driver1) &&
                        (FrmBoca.Current_Driver_Type != FrmBoca.Boca_Spooler_Driver3))   //5.2.0.0
                    {
                        if (sn == null)
                            pn = pn + ", Port";                         //build port name
                        else
                            pn = sn + backslash + pn + ", Port";        //build port name with server name
                    }
                    //End of test for 5.1.0.0
                    // Free memory
                    Marshal.FreeCoTaskMem(pBytes);
                }
                ClosePrinter(hPrinter);
            }

            //return the port name for use in attempting to read from the printer
            return pn;
        }

        //1.0.5.0 check the printer attributes to see if printer online or offline
        public static bool Is_Printer_Online()
        {

            bool online = false;            //initialize to false
            int att = 0;
            Int32 level = 2;
            Int32 Needed = 0;
            IntPtr pPrinter = new IntPtr(0);
            PRINTER_INFO_2 pi = new PRINTER_INFO_2();
            PRINTER_INFO_6 pi6 = new PRINTER_INFO_6();
            byte[] pBuffer = null;
            pBuffer = new byte[1024];

            // Get printer level page size needed - first call
            if (GetPrinter(Write_Printer_Handle, level, pPrinter, 0, out Needed) == false)
            {
                if (Needed > 0)
                {
                    // Allocate needed memory
                    IntPtr pBytes = Marshal.AllocCoTaskMem(Needed);

                    // Get printer level data - second call
                    GetPrinter(Write_Printer_Handle, level, pBytes, Needed, out Needed);

                    // Convert printer data block into class structure
                    Marshal.PtrToStructure(pBytes, pi);
                    att = pi.Attributes;

                    //Save value read into global variable Printer_Info_2_Status
                    FrmBoca.Printer_Info_2_Status = pi.Status;

                    // Free memory
                    Marshal.FreeCoTaskMem(pBytes);

                    //5.4.0.0 If using Bidi printer monitor, check the Printer_Info structure maintained by the monitor 
                    //to determine online/offline status
                    if ((boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver1) ||
                        (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver3) ||
                        (boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Standard_Driver1))   //5.2.0.0
                    {
                        if ((FrmBoca.Printer_Info_2_Status & PRINTER_STATUS_OFFLINE) == 0)
                            online = true;
                        else
                            online = false;
                    }
                    else    //5.4.0.0
                    {
                        //check if printer online using the attributes field when not using printer monitor
                        if ((att & PRINTER_ATTRIBUTE_WORK_OFFLINE) == 0)
                            online = true;
                    }

                    //Check Printer_Info_6 for further status
                    Needed = 0;
                    level = 6;

                    // Get printer level 6 page size needed - first call
                    if (GetPrinter(Write_Printer_Handle, level, pPrinter, 0, out Needed) == false)
                    {
                        if (Needed > 0)
                        {
                            // Allocate needed memory
                            pBytes = Marshal.AllocCoTaskMem(Needed);

                            // Get printer level 6 data - second call
                            GetPrinter(Write_Printer_Handle, level, pBytes, Needed, out Needed);

                            // Convert printer data block into class structure
                            Marshal.PtrToStructure(pBytes, pi6);

                            //Save value read into global variable
                            FrmBoca.Printer_Info_6_dwStatus = pi6.dwStatus;

                            // Free memory
                            Marshal.FreeCoTaskMem(pBytes);
                        }
                    }
                }
            }

            //Check if offline if not using monitor, close any open channels          //5.4.0.0
            if (!online)
            {
                if (!(boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver1) &&
                    !(boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver3) &&
                    !(boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Standard_Driver1))   //5.2.0.0
                    quick_close();
            }

            return (online);
        }

        public void readDataThread()
        {
            bool online = false;

            //loop as long as checking for data is needed.  When executing Close_Printer the
            //boolean variable checking_for_data is set to false, and the read thread is aborted.
            while (boca_sample.FrmBoca.checking_for_data)
            {
                //1.0.5.0 check if printer still online...
                online = Is_Printer_Online();                                    //if offline, close channels

                //if online attempt the read
                if (online)                                                      //1.0.5.0 
                {
                    m_parent.incoming = Read_Printer();                          //perform read
                    if (m_parent.incoming.Length > 0)                            //if read successful use callback delegate
                        m_parent.Establish_Status();                             //to relay status back to main thread
                }
                else                                                             //1.0.5.0 
                {
                    //If offline and not using monitor, show closing of any open channels          //5.4.0.0
                    if (!(boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver1) &&
                        !(boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Spooler_Driver3) &&
                        !(boca_sample.FrmBoca.Current_Driver_Type == boca_sample.FrmBoca.Boca_Standard_Driver1))   //5.4.0.0
                        m_parent.spo();                                          //1.1.2.1 Show printer offline
                }
            }
            m_parent.monitoring = false;                                         //5.0.0.0
        }               //end of read data thread

        //This routine is used to control the read thread states.
        //from here a read thread can be created, started and stopped.
        public void readData()
        {

            if (dataReadingThread.ThreadState.ToString() == "Unstarted")
            {
                //start the thread
                dataReadingThread.Start();
                Thread.Sleep(0);
            }
            else if (dataReadingThread.ThreadState.ToString() == "Running")
            {
                //end enternal loop of reading printer status
                boca_sample.FrmBoca.checking_for_data = false;

                //Without overlapped IO the read data thread is still sitting at the 
                //ReadPrinter statement in Read_Printer.
                //Until overlapped IO is working on the USB port it is necessary to send
                //an <S1> to the printer to generate return status XOn.  The read data thread
                //will read the XOn and then exit.
                Write_Printer("<S1>");

                //Stop the Thread
                dataReadingThread.Abort();
                Thread.Sleep(0);
            }
            else if (dataReadingThread.ThreadState.ToString() == "AbortRequested")  //1.0.5.0
            {
                //@mdh test aborting read thread, need some code here
                //MessageBox.Show("Aborting Read thread...");
            }
            else if (dataReadingThread.ThreadState.ToString() == "Stopped")         //1.0.5.0
            {
                //@mdh test aborting read thread, need some code here
                //MessageBox.Show("Read thread Stopped...");
            }
            else
            {
                //create Read Thread
                dataReadingThread = new Thread(new ThreadStart(readDataThread));
                Thread.Sleep(0);

                boca_sample.FrmBoca.checking_for_data = true;

                //start the thread
                dataReadingThread.Start();
                Thread.Sleep(0);

            }

        }

        public static void start_stop_read_thread()
        {
            //5.1.0.0
            if (boca_sample.FrmBoca.Current_Driver_Type != boca_sample.FrmBoca.Non_Boca_Driver)
            {
                //Start/Stop read data thread
                if (boca_raw != null)                   //1.0.5.0
                    boca_raw.readData();
            }

        }

        // Disable printer (offline).
        static unsafe void CSDisablePrinter()
        {

            IntPtr pBytes = Marshal.AllocCoTaskMem(PRINTER_STATUS_OFFLINE);
            if (!SetPrinter(Write_Printer_Handle, 0, pBytes, PRINTER_CONTROL_SET_STATUS))
            {
                throw new ApplicationException("Cannot disable printer " + Marshal.GetLastWin32Error());
            }
            Marshal.FreeCoTaskMem(pBytes);
        }

        // Enable printer.
        static unsafe void CSEnablePrinter()        //(int hPrinter)
        {
            IntPtr pBytes = Marshal.AllocCoTaskMem(PRINTER_STATUS_READY);
            if (!SetPrinter(Write_Printer_Handle, 0, pBytes, PRINTER_CONTROL_SET_STATUS))
            {
                throw new ApplicationException("Cannot enable printer " + Marshal.GetLastWin32Error());
            }
            Marshal.FreeCoTaskMem(pBytes);
        }

        //7.0.0.0  SetPrinterSettings
        public static bool SetPrinterSettings(string FormName, short PaperSize, short PaperSource, short PaperOrientation, short PaperCopies)
        {
            bool bFlag = true;

            Int32 level = 2;
            Int32 Needed = 0;
            IntPtr hPrinter = new IntPtr(0);
            IntPtr pPrinter = new IntPtr(0);

            PRINTER_INFO_2 pi = new PRINTER_INFO_2();
            structDevMode devMode = new structDevMode();
            string szPrinterName = FrmBoca.PrinterDriverName;

            PRINTERDEFAULTS df = new PRINTERDEFAULTS();

            //Init Printer Default Structure
            df.pDataType = "RAW";
            df.pDevMode = 0;

            if ((FrmBoca.CurrentOperatingSystem == "Windows 8/10") || (FrmBoca.CurrentOperatingSystem == "Windows 10"))           //test win 10
                df.DesiredAccess = PRINTER_ACCESS_MANAGE_LIMITED;           //win 10
            else
                df.DesiredAccess = PRINTER_ALL_ACCESS;                      //win           

            //setup int pointer to printer default structure
            IntPtr def = Marshal.AllocHGlobal(Marshal.SizeOf(df));
            Marshal.StructureToPtr(df, def, false);

            // Open the printer.
            if (OpenPrinter(szPrinterName.Normalize(), out hPrinter, IntPtr.Zero))
            {
                // Get printer level page size needed - first call
                if (GetPrinter(hPrinter, level, pPrinter, 0, out Needed) == false)
                {
                    // Allocate needed memory
                    IntPtr pBytes = Marshal.AllocCoTaskMem(Needed);

                    // Get printer level data - second call
                    GetPrinter(hPrinter, level, pBytes, Needed, out Needed);

                    // GET THE SIZE OF THE DEV_MODE BUFFER
                    int dwNeeded = DocumentProperties(IntPtr.Zero, hPrinter, szPrinterName, IntPtr.Zero, IntPtr.Zero, 0);
                    IntPtr pDevMode = Marshal.AllocHGlobal(dwNeeded);

                    // Convert printer data block into class structure
                    pi = (PRINTER_INFO_2)Marshal.PtrToStructure(pBytes, pi.GetType());

                    // Convert printer data block into class structure
                    Marshal.PtrToStructure(pBytes, pi);

                    // FILL THE DEV_MODE STRUCTURE
                    devMode = (structDevMode)Marshal.PtrToStructure(pi.pDevMode, devMode.GetType());

                    // SET THE FORM NAME FIELDS TO INDICATE THAT THIS FIELD WILL BE MODIFIED
                    //devMode.dmFields |= devMode.dmFields & DM_FORMNAME & DM_PAPERSIZE & DM_ORIENTATION & DM_COPIES;
                    devMode.dmFields |= DM_FORMNAME;
                    devMode.dmFields |= DM_PAPERSIZE;
                    devMode.dmFields |= DM_DEFAULTSOURCE;
                    devMode.dmFields |= DM_ORIENTATION;
                    devMode.dmFields |= DM_COPIES;
                    
                    //Set values picked by user
                    devMode.dmFormName = FormName;
                    devMode.dmPaperSize = PaperSize;
                    devMode.dmDefaultSource = PaperSource;
                    devMode.dmOrientation = PaperOrientation;
                    devMode.dmCopies = PaperCopies;
                    Marshal.StructureToPtr(devMode, pDevMode, true);

                    //Modify PRINTER_INFO_2 devmode
                    int iRet = DocumentProperties(IntPtr.Zero, hPrinter, szPrinterName, pi.pDevMode, pDevMode, DM_IN_BUFFER | DM_OUT_BUFFER);
                    if (iRet < 0)
                    {
                        MessageBox.Show("Unable to set the Document properties for this printer." + Marshal.GetLastWin32Error());
                        bFlag = false;
                    }
                    else
                    {
                        //Set pointer
                        Marshal.PtrToStructure(pBytes, pi);

                        //Save Printer Driver settings
                        if (!SetPrinter(hPrinter, level, pBytes, 0))
                        {
                            MessageBox.Show("Cannot save print driver properties " + Marshal.GetLastWin32Error());
                            bFlag = false;
                        }
                    }

                    // Free memory
                    Marshal.FreeCoTaskMem(pBytes);
                    //Marshal.FreeHGlobal(pDevMode);
                }
                else
                {
                    MessageBox.Show("Cannot Get Printer Data " + Marshal.GetLastWin32Error());
                    bFlag = false;
                }
                ClosePrinter(hPrinter);
            }
            else
            {
                MessageBox.Show("Cannot Open Printer " + Marshal.GetLastWin32Error());
                bFlag = false;
            }

            return bFlag;

        }

        public static void PDF_Print(string file_name)
        {
            char doublequote = (char)34;
            char backslash = (char)92;
            string batchfile = "";

            //test for bitness
            if (IntPtr.Size == 8)                                                                    
                batchfile = @"C:\Program Files (x86)\Adobe\Acrobat Reader DC\Reader\AcroRd32.exe";   //for 32 bit operating systems
            else
                batchfile = @"C:\Program Files\Adobe\Acrobat Reader DC\Reader\AcroRd32.exe";         //for 64 bit operating systems

            //Start new process to spawn
            Process p = new Process();
            //Run AcroRd32.exe
            p.StartInfo.FileName = batchfile;
            // /N => Start new AcroRd32.exe session each time and when complete kill all sessions but one
            // /h => hide window
            // /t => send FILENAME to PRINTERNAME
            p.StartInfo.Arguments = " /N /h /t " + doublequote + file_name + doublequote + " " + doublequote + FrmBoca.PrinterDriverName + doublequote;
            //No window displayed
            p.StartInfo.CreateNoWindow = true;
            //Hide Window
            p.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
            //have adobe go staraight to print function
            p.StartInfo.Verb = "print";
            p.StartInfo.RedirectStandardError = false;
            p.StartInfo.RedirectStandardOutput = false;
            p.StartInfo.UseShellExecute = false;
            p.StartInfo.WorkingDirectory = FrmBoca.parent_dir + backslash;
            p.Start();

            //If you remove /N Arguement above, use code below to shut down AcroRd32.exe
            //System.Threading.Thread.Sleep(4000);
            //if (!p.CloseMainWindow())
            //p.Kill();

        }
        //For PDF
        //See https://social.msdn.microsoft.com/Forums/office/en-US/2567a06f-c3d9-4d99-a0c9-824cc6a23705/printing-a-pdf-file-to-a-printer?forum=vcgeneral 
        //See also https://www.codeproject.com/Tips/598424/How-to-Silently-Print-PDFs-using-Adobe-Reader-and
        //And https://msdn.microsoft.com/en-us/library/windows/desktop/dd145226(v=vs.85).aspx
        //and https://stackoverflow.com/questions/17448465/send-pdf-file-to-a-printer-print-pdf


        //7.0.0.0 added to change between three known driver modes
        //Mode #1 - Spool Unidirectional
        //Mode #2 - Direct Bidirectional
        //Mode #3 - Spool Bidirectional
        public static void Set_Attributes(string szPrinterName, int mode)
        {
            string filename = "";
            char doublequote = (char)34;
            char backslash = (char)92;
            //string test = "";

            //Establish driver mode before opening printer ports
            switch (mode)
            {
                //Driver Mode #3
                case 3:
                    filename = "mode3.bat";                 //Set Spool Mode On and BIDI Support On
                    break;
                //Driver Mode #2
                case 2:
                    filename = "mode2.bat";                 //Set Direct Mode On and BIDI Support Off
                    break;
                //Driver Mode #1
                default:
                    filename = "mode1.bat";                 //Set Spool Mode On and BIDI Support Off
                    break;
            }

            System.Diagnostics.Process p = new System.Diagnostics.Process();
            p.StartInfo.RedirectStandardError = false;
            p.StartInfo.RedirectStandardOutput = false;
            p.StartInfo.UseShellExecute = false;
            p.StartInfo.WorkingDirectory = FrmBoca.parent_dir + backslash;
            p.StartInfo.FileName = FrmBoca.parent_dir + backslash + filename;
            p.StartInfo.CreateNoWindow = true;
            p.StartInfo.Arguments = " " + doublequote + szPrinterName + doublequote;
            p.Start();
            Thread.Sleep(1000);
            p.WaitForExit();

        }

    }

}
