﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

//7.0.0.0 Set Default Printer Driver settings
namespace boca_sample
{
    using System.Drawing.Printing;

    public partial class FrmPref : Form
    {
        public const short DMORIENT_PORTRAIT = 1;
        public const short DMORIENT_LANDSCAPE = 2;

        public static PrintDialog pd;                       //printer driver dialog
        public string FormName;
        public int PaperSize;
        public PaperSource Papersource;
        public bool PaperOrientation;
        public int PaperCopies;
        private FrmBoca m_parent;

        public FrmPref(FrmBoca frm1)
        {
            InitializeComponent();
            m_parent = frm1;
            PopulateDriverOptions();
        }

        //7.0.0.0 PopulateDriverOptions
        public void PopulateDriverOptions()
        {
            PaperSource ps;
            PaperSize pkSize;
            int x = 0;
            int y = 0;
            int z = 0;
            string MenuItem;
            bool result;

            pd = new PrintDialog();
            pd.PrinterSettings = new PrinterSettings();
            pd = FrmBoca.pd;

            comboPaperSource.Items.Clear();
            comboPaperSize.Items.Clear();
            
            // Set default printer
            if (!pd.PrinterSettings.IsDefaultPrinter)
            {
                result = RawPrinterHelper.SetDefaultPrinter(pd.PrinterSettings.PrinterName.Normalize());
                //MessageBox.Show("Result = " + result.ToString());
            }
            CurrentDriver.Text = pd.PrinterSettings.PrinterName;

            //Read and display currently selected driver settings for source, size, copies and orientation
            for (int i = 0; i < pd.PrinterSettings.PaperSources.Count; i++)
            {
                //Parse PaperSource Items for dropdown menu
                ps = pd.PrinterSettings.PaperSources[i];
                x = ps.ToString().IndexOf(" ", 0);
                y = ps.ToString().IndexOf(" Kind", x + 1);
                z = y - (x + 1);
                MenuItem = ps.ToString().Substring(x + 1, z);
                comboPaperSource.Items.Add(MenuItem);
            }

            for (int i = 0; i < pd.PrinterSettings.PaperSizes.Count; i++)
            {
                //Parse PaperSize Items for dropdown menu
                pkSize = pd.PrinterSettings.PaperSizes[i];
                x = pkSize.ToString().IndexOf(" ", 0);
                y = pkSize.ToString().IndexOf(" Kind", x + 1);
                z = y - (x + 1);
                MenuItem = pkSize.ToString().Substring(x + 1, z);
                comboPaperSize.Items.Add(MenuItem);
            }
            //Include current default prefrences in global variables
            FormName = pd.PrinterSettings.DefaultPageSettings.PaperSize.PaperName;
            PaperSize = pd.PrinterSettings.DefaultPageSettings.PaperSize.RawKind;
            Papersource = pd.PrinterSettings.DefaultPageSettings.PaperSource;
            PaperOrientation = pd.PrinterSettings.DefaultPageSettings.Landscape;
            PaperCopies = pd.PrinterSettings.DefaultPageSettings.PrinterSettings.Copies;

            //Display values on screen
            Display_Settings();
        }

        public void Display_Settings()
        {

            PaperSource ps;
            PaperSize pkSize;
            int x = 0;
            int y = 0;
            int z = 0;
            string MenuItem;
            
            //Use current default setting and display thru radio button settings
            if (pd.PrinterSettings.DefaultPageSettings.Landscape == true)
                OptLandscape.Checked = true;
            else
                OptPortrait.Checked = true;

            //Use current default setting to display copies count
            CopyCount.Value = pd.PrinterSettings.DefaultPageSettings.PrinterSettings.Copies;

            //Parse current default setting to display PaperSource
            ps = pd.PrinterSettings.DefaultPageSettings.PaperSource;
            x = ps.ToString().IndexOf(" ", 0);
            y = ps.ToString().IndexOf(" Kind", x + 1);
            z = y - (x + 1);
            MenuItem = ps.ToString().Substring(x + 1, z);
            CurrentSource.Text = MenuItem;

            //Parse current default setting to display PaperSize
            pkSize = pd.PrinterSettings.DefaultPageSettings.PaperSize;
            x = pkSize.ToString().IndexOf(" ", 0);
            y = pkSize.ToString().IndexOf(" Kind", x + 1);
            z = y - (x + 1);
            MenuItem = pkSize.ToString().Substring(x + 1, z);
            CurrentSize.Text = MenuItem;


        }
        private void OptPortrait_CheckedChanged(object sender, EventArgs e)
        {
            pd.PrinterSettings.DefaultPageSettings.Landscape = false;
            PaperOrientation = false;
        }

        private void OptLandscape_CheckedChanged(object sender, EventArgs e)
        {
            pd.PrinterSettings.DefaultPageSettings.Landscape = true;
            PaperOrientation = true;
        }

        private void comboPaperSource_SelectedIndexChanged(object sender, EventArgs e)
        {
            pd.PrinterSettings.DefaultPageSettings.PaperSource = pd.PrinterSettings.PaperSources[comboPaperSource.SelectedIndex];
            Papersource = pd.PrinterSettings.PaperSources[comboPaperSource.SelectedIndex];
        }

        private void comboPaperSize_SelectedIndexChanged(object sender, EventArgs e)
        {
            PrinterSettings ps = new PrinterSettings();
            PrintDocument pdoc = new PrintDocument();
            //string Name;
            //Int32 Height, Width;

            pd.PrinterSettings.DefaultPageSettings.PaperSize = pd.PrinterSettings.PaperSizes[comboPaperSize.SelectedIndex];

           // Name = pd.PrinterSettings.DefaultPageSettings.PaperSize.PaperName;
            //Height = pd.PrinterSettings.DefaultPageSettings.PaperSize.Height;
            //Width = pd.PrinterSettings.DefaultPageSettings.PaperSize.Width;
            //pdoc.DefaultPageSettings.PaperSize = new System.Drawing.Printing.PaperSize(Name, Width, Height);
            
            FormName = pd.PrinterSettings.DefaultPageSettings.PaperSize.PaperName;
            PaperSize = pd.PrinterSettings.PaperSizes[comboPaperSize.SelectedIndex].RawKind;

        }

        private void CopyCount_ValueChanged(object sender, EventArgs e)
        {
            PaperCopies = Convert.ToInt32(CopyCount.Value);
            pd.PrinterSettings.Copies = (short)PaperCopies;
        }
 
        private void CmdOK_Click(object sender, EventArgs e)
        {

            short orientation;

            if (PaperOrientation)
                orientation = DMORIENT_LANDSCAPE;
            else
                orientation = DMORIENT_PORTRAIT;

            //Set operating system variables
            bool status = RawPrinterHelper.SetPrinterSettings(FormName, (short)PaperSize, (short)Papersource.RawKind, orientation, (short)PaperCopies);
            //Close open window
            this.Close();
            Display_Settings();
        }

        private void CmdApply_Click(object sender, EventArgs e)
        {

            short orientation;

            if (PaperOrientation)
                orientation = DMORIENT_LANDSCAPE;
            else
                orientation = DMORIENT_PORTRAIT;

            //Set operating system variables
            bool status = RawPrinterHelper.SetPrinterSettings(FormName, (short)PaperSize, (short)Papersource.RawKind, orientation, (short)PaperCopies);
            //Leave Window open
            Display_Settings();
        }

        private void CmdCancel_Click(object sender, EventArgs e)
        {
            Display_Settings();
            //Close window without saving OS variables
            this.Close();
        }
    }
}
