﻿//
// WiFi Setup routines
//
//
// SSID - 32 chars max used with FGL command <wid#>SSID (where # is length of SSID)
//
// Enable / Disable wireless unit with FGL command <wnm#>
//
//    <wnm0> - Disable Wireless
//    <wnm1> - Enable with Static IP
//    <wnm2> - Enable with DHCP IP
//
// Wireless Security Modes use FGL command <wkt#> where # is key type
//
//  0. None - No security selected
//  1. WEP - will transmit 13 bytes per key code.  If method #a below then pad
//           with zeroes. Use FGL command <wkd#>Key where # is length.
//      a.  64 Bit - 10 digits for hex code (5 bytes)
//      b. 128 Bit - 26 digits for hex code (13 bytes)
//  2. WPA  - will transmit "Pass Phrase" 8 - 63 chars long. Use FGL
//            command <wpp#>Phrase to pass the phrase to the printer.
//  3. WPA2 - will transmit "Pass Phrase" 8 - 63 chars long. Use FGL
//            command <wpp#>Phrase to pass the phrase to the printer.
//
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace boca_sample
{

    public partial class FrmWiFi : Form
    {
        const int DISABLED = 0;
        const int WEP = 1;
        const int WPA = 2;
        const int WPA2 = 3;

        int def_transmit_key, key_type, max_key_length;
        string wireless_command;
        private FrmBoca m_parent;

        public FrmWiFi(FrmBoca frm1)
        {
            InitializeComponent();
            m_parent = frm1;
        }

        private void FrmWiFi_Load(object sender, EventArgs e)
        {
            security.Items.Add("Disabled");
            security.Items.Add("WPA");
            security.Items.Add("WPA2");
            security.Items.Add("WEP");
            wep_encryption.Items.Add("64 Bits");
            wep_encryption.Items.Add("128 Bits");
            wireless_mode.Items.Add("Disable Wireless");
            wireless_mode.Items.Add("Enable with Static IP");
            wireless_mode.Items.Add("Enable with DHCP IP");
            initialize_form();
            ssid.Visible = false;
            ssid_label.Visible = false;
            security.Visible = false;

        }
        private void initialize_form()
        {
            wep_encryption.Visible = false;
            DTK_Frame.Visible = false;
            DTK_label.Visible = false;
            pass_phrase_label.Visible = false;
            pass_phrase.Visible = false;
            key1.Visible = false;
            key1_label.Visible = false;
            key2.Visible = false;
            key2_label.Visible = false;
            key3.Visible = false;
            key3_label.Visible = false;
            key4.Visible = false;
            key4_label.Visible = false;
            CmdOk.Visible = false;
            option_label.Visible = false;
           
        }

        private void initialize_values()
        {
            key1.Text = "";
            key2.Text = "";
            key3.Text = "";
            key4.Text = "";
            pass_phrase.Text = "";
            wep_encryption.Text = "WEP Encryption";
            option1.Checked = true;
            def_transmit_key = 0;
        }

        private void key_on()
        {
            if (security.Text == "WEP")
            {
                key1_label.Visible = true;
                key1.Visible = true;
                key2_label.Visible = true;
                key2.Visible = true;
                key3_label.Visible = true;
                key3.Visible = true;
                key4_label.Visible = true;
                key4.Visible = true;
                option_label.Visible = false;
            }
            else
            {
                key1_label.Visible = true;
                key1.Visible = true;
                option_label.Visible = true;
            }
        }

        private void key_off()
        {
            key1_label.Visible = false;
            key1.Visible = false;
            key2_label.Visible = false;
            key2.Visible = false;
            key3_label.Visible = false;
            key3.Visible = false;
            key4_label.Visible = false;
            key4.Visible = false;
        }

        //verify hexidecimal character is 0 to 9, a to f, or A to F
        private bool validate_char(char character)
        {
            bool status;

            if (((character >= 48) && (character <= 57)) ||
                ((character >= 65) && (character <= 70)) ||
                ((character >= 97) && (character <= 102)))
                status = true;
            else
                status = false;

            return (status);
        }

        //<wid#>ssid
        private void send_wid()
        {
            m_parent.Transmit("<wid" + ssid.TextLength + ">" + ssid.Text);
        }

        //<wkt#>security mode
        //0 - None/Disabled
        //1 - WEP
        //2 - WPA
        //3 - WPA2
        private void send_wkt()
        {
            m_parent.Transmit("<wkt" + key_type.ToString() + ">");
        }

        //<wpp#>pass_phrase
        private void send_wpp()
        {
            m_parent.Transmit("<wpp" + pass_phrase.TextLength + ">" + pass_phrase.Text);
        }

        //<wwi#>Default Transmit Key
        //Option value for default transmit key
        //1 - Value 0
        //2 - Value 1
        //3 - Value 2
        //4 - Value 3
        private void send_wwi()
        {
            m_parent.Transmit("<wwi" + def_transmit_key + ">");
        }

        //<wkd#>key_code
        //read each two byte hex number from string and convert to decimal
        private void send_wkd()
        {
            string incoming, outgoing, str;
            int i, j, x;
            string zero;

            i = 0;
            zero = i.ToString();
            outgoing = "";

            //if security mode is enabled as WPA or WPA2 send only key code #1
            if (((key_type == WPA) || (key_type == WPA2)) && (key1.TextLength > 0))
            {
                incoming = key1.Text;
                for (i = 0; i < incoming.Length; i++)
                {
                    //strip off two bytes from string
                    str = incoming.Substring(i, 2);
                    //convert from hex to dec
                    x = Convert.ToInt32(str, 16);
                    //build outgoing string from chr(x)
                    outgoing += (char)x;
                    i++;
                }
                i = incoming.Length / 2;

                //generate final outgoing string and transmit it to the printer
                outgoing = "<wpk" + i.ToString() + ">" + outgoing;

            }

            //if however security mode is WEP, send key codes #1 thru #4.  The WEP encryption
            //can be 64 bit or 128 bit.  This result in 10 or 26 digit key code.  To simplify this
            //for the printer, all codes will be transmitted as a 26 digit key code by padding 16
            //zeros on the end of the 10 digit code
            if (key_type == WEP)
            {
                //max key length will be 10 or 26 in WEP mode
                if (max_key_length == 10)
                    outgoing = "<wke0><wpk52>";
                else
                    outgoing = "<wke1><wpk52>";

                incoming = "";
                for (j = 0; j < 4; j++)
                {
                    switch (j)
                    {
                        case 0:
                            incoming = key1.Text;
                            break;
                        case 1:
                            incoming = key2.Text;
                            break;
                        case 2:
                            incoming = key3.Text;
                            break;
                        case 3:
                            incoming = key4.Text;
                            break;
                    }
                    if (incoming.Length < 26)
                    {
                        x = 26 - incoming.Length;
                        for (i = 0; i < x; i++)
                            incoming += zero;
                    }

                    for (i = 0; i < 26; i++)
                    {
                        //strip off two bytes from string
                        str = incoming.Substring(i, 2);
                        //convert from hex to dec
                        x = Convert.ToInt32(str, 16);
                        //build outgoing string from chr(x)
                        outgoing += (char)x;
                        i++;
                    }
                }               //end for j
            }                   //endif (key_type == WEP)

            //if there is an outgoing string made, then transmit it
            if (outgoing.Length > 0)
            {
                byte[] buffer = new byte[outgoing.Length];
                buffer = System.Text.Encoding.ASCII.GetBytes(outgoing);

                m_parent.Transmitb(buffer);

             }

        }                       //end send_wkd()

        //Enable/Disable Wireless Mode
        private void wireless_mode_SelectedIndexChanged(object sender, EventArgs e)
        {
            ssid.Text = "";
            security.Text = "Security Mode";
            security.Visible = false;
            wireless_command = "<wnm" + wireless_mode.SelectedIndex + ">";
            if (wireless_mode.SelectedIndex == 0)
            {        
                ssid.Visible = false;
                ssid_label.Visible = false;
                CmdOk.Visible = true;
                pass_phrase_label.Visible = false;
                pass_phrase.Visible = false;
                max_key_length = 0;
                key_type = DISABLED;        //Disabled security mode
                CmdOk.Focus();
            }
            else
            {
                CmdOk.Visible = false;
                ssid.Visible = true;
                ssid_label.Visible = true;
                ssid.Focus();
            }
        }

        private void ssid_TextChanged(object sender, EventArgs e)
        {
            string s = ssid.Text;
            security.Text = "Security Mode";
            initialize_values();
            initialize_form();
            if (s.Length > 32)
            {
                m_parent.Display_Text("SSID maximum length of 32 characters has been exceeded");
                ssid.Text = "";
                security.Visible = false;
            }
            else
                security.Visible = true;
        }

        private void security_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (ssid.Text == "")
            {
                m_parent.Display_Text("You must enter a Wireless Network name above (SSID)");
                security.Text = "Security Mode";
                ssid.Focus();
            }
            else
            {
                initialize_values();
                initialize_form();
                if (security.Text == "Disabled")
                {
                    CmdOk.Visible = true;
                    pass_phrase_label.Visible = false;
                    pass_phrase.Visible = false;
                    max_key_length = 0;
                    key_type = DISABLED;               //Disabled security mode
                    CmdOk.Focus();
                }
                else
                {
                    if (security.Text == "WPA")
                    {
                        pass_phrase_label.Visible = true;
                        pass_phrase.Visible = true;
                        max_key_length = 64;
                        key_type = WPA;                    //WPA security mode
                        pass_phrase.Focus();
                    }
                    else
                    {
                        if (security.Text == "WPA2")
                        {
                            pass_phrase_label.Visible = true;
                            pass_phrase.Visible = true;
                            max_key_length = 64;
                            key_type = WPA2;                //WPA2 security mode
                            pass_phrase.Focus();
                        }
                        else    				//Must be WEP Mode
                        {
                            pass_phrase_label.Visible = true;
                            wep_encryption.Visible = true;
                            key_type = WEP;                //WEP security mode
                            wep_encryption.Focus();
                        }
                    }
                }
            }
        }                       //end of security change

        private void wep_encryption_SelectedIndexChanged(object sender, EventArgs e)
        {
            
            if (wep_encryption.Text == "64 Bits")
                max_key_length = 10;
            else
                max_key_length = 26;
            pass_phrase_label.Visible = true;
            pass_phrase.Visible = true;
            DTK_Frame.Visible = true;
            DTK_label.Visible = true;
            pass_phrase.Focus();
        }

        private void pass_phrase_TextChanged(object sender, EventArgs e)
        {
            int x = pass_phrase.TextLength;
                
            if (x > 63)
            {
                if (security.Text == "WEP")
                    m_parent.Display_Text("WEP Pass Phrase maximum length of 63 characters has been exceeded");
                else
                    m_parent.Display_Text("WPA Shared Key maximum length of 63 characters has been exceeded");

                pass_phrase.Text = "";
                CmdOk.Visible = false;
                key_off();
            }
            else
            {
                if (x >= 8)
                {
                    key_on();
                    CmdOk.Visible = true;
                }
                else
                {
                    key_off();
                    CmdOk.Visible = false;
                }
            }
        }

        private void option1_CheckedChanged(object sender, EventArgs e)
        {
            def_transmit_key = 0;
        }

        private void option2_CheckedChanged(object sender, EventArgs e)
        {
            def_transmit_key = 1;
        }                           //end class

        private void option3_CheckedChanged(object sender, EventArgs e)
        {
            def_transmit_key = 2;
        }

        private void option4_CheckedChanged(object sender, EventArgs e)
        {
            def_transmit_key = 3;
        }

        private void key1_TextChanged(object sender, EventArgs e)
        {
            bool success;
            char c;
            string s, incoming = key1.Text;
            int x = incoming.Length;

            if (x > max_key_length)
            {
                m_parent.Display_Text("Key field length of " + max_key_length.ToString() + " has been exceeded");
                key1.Text = "";
            }
            else
            {
                s = incoming.Substring(x-1, 1);
                c = s[0];
                success = validate_char(c);
                if (success == false)
                {
                    if (x > 0)
                        m_parent.Display_Text("Invalid Hex character entered " + c + ".  Hex character must be 0 to 9, or A to F");
                    key1.Text = incoming.Substring(0, x - 1);
                }
            }
        }

        private void key2_TextChanged(object sender, EventArgs e)
        {
            bool success;
            char c;
            string s, incoming = key2.Text;
            int x = incoming.Length;

            if (x > max_key_length)
            {
                m_parent.Display_Text("Key field length of " + max_key_length.ToString() + " has been exceeded");
                key2.Text = "";
            }
            else
            {
                s = incoming.Substring(x-1, 1);
                c = s[0];
                success = validate_char(c);
                if (success == false)
                {
                    if (x > 0)
                        m_parent.Display_Text("Invalid Hex character entered " + c + ".  Hex character must be 0 to 9, or A to F");
                    key2.Text = incoming.Substring(0, x - 1);
                }
            }
        }

        private void key3_TextChanged(object sender, EventArgs e)
        {
            bool success;
            char c;
            string s, incoming = key3.Text;
            int x = incoming.Length;

            if (x > max_key_length)
            {
                m_parent.Display_Text("Key field length of " + max_key_length.ToString() + " has been exceeded");
                key3.Text = "";
            }
            else
            {
                s = incoming.Substring(x-1, 1);
                c = s[0];
                success = validate_char(c);
                if (success == false)
                {
                    if (x > 0)
                        m_parent.Display_Text("Invalid Hex character entered " + c + ".  Hex character must be 0 to 9, or A to F");
                    key3.Text = incoming.Substring(0, x - 1);
                }
            }
        }

        private void key4_TextChanged(object sender, EventArgs e)
        {
            bool success;
            char c;
            string s, incoming = key4.Text;
            int x = incoming.Length;

            if (x > max_key_length)
            {
                m_parent.Display_Text("Key field length of " + max_key_length.ToString() + " has been exceeded");
                key4.Text = "";
            }
            else
            {
                s = incoming.Substring(x-1, 1);
                c = s[0];
                success = validate_char(c);
                if (success == false)
                {
                    if (x > 0)
                        m_parent.Display_Text("Invalid Hex character entered " + c + ".  Hex character must be 0 to 9, or A to F");
                    key4.Text = incoming.Substring(0, x - 1);
                }
            }
        }

        private void CmdOk_Click(object sender, EventArgs e)
        {

            //enable FGL command processor in HP printer and enable buffered commands in printer
            m_parent.Transmit("<VS1><bce>");
            m_parent.Transmit(wireless_command);       //Enable / Disable wireless communication

            //If wireless communication enabled send wireless security commands
            if (wireless_mode.SelectedIndex > 0)
            {

                send_wid();             //send the SSID to printer using <wid#>
    
                //if security mode is enabled as WPA or WPA2 check length of key code #1
                if ((key_type == WPA) || (key_type == WPA2))
                {
                    if (key1.TextLength > 0)
                        key_type = WPA;
                    else
                        key_type = WPA2;
                }

                send_wkt();                    //send the Key Type to printer using <wkt#>
    
                if (security.Text != "Disabled")
                {
                    if (key_type == WEP)
                        send_wwi();                //Send default transmit key value 0 -3
                    send_wpp();                    //send the pass phrase to printer using <wpp#>
                    send_wkd();                    //send 1 to 4 key codes to printer using <wkd#>
                }
                m_parent.Transmit("<bcd>");             //disable buffered commands in the printer
            }
                        
        }

    }                           //end class
}                               //end namespace
